/*==================================================================
  fft.c

  Copyright (c) 2004-2006 N099
==================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "fft.h"

/*------------------------------------------------------------------
  internal utility functions
------------------------------------------------------------------*/
static int make_bit_reverse_tbl(int, int*);
static void make_coeff_tbl(int, int, Complex *);
static void free0(void *);
static int fft_1D2(int num, int s, Complex *, int t, int end);

/*------------------------------------------------------------------
  1D,2D memory allocation
------------------------------------------------------------------*/
Complex *mallocComplex1D(int num)
{
  if(num <= 0){
    return NULL;
  }
  
  num = check_2spow(num);
  return calloc(num, sizeof(Complex));
}

Complex **mallocComplex2D(int col, int row)
{
  int i;
  void *p;
  Complex *pc;
  Complex **ppc;
  
  if(col <= 0 || row <= 0){
    return NULL;
  }
  
  col = check_2spow(col);
  row = check_2spow(row);

  //ꂾƃACg
  p = calloc(col * row * sizeof(Complex) + col * sizeof(Complex*), 1);
  if(NULL == p){
    return NULL;
  }

  ppc = p;
  pc = (Complex *)&ppc[col];
  for(i=0; i<col; ++i){
    ppc[i] = pc;
    pc += row;
  }
  
  return p;
}
  

void freeComplex1D(Complex *c)
{
  if(c!=NULL){
    free(c);
  }
}

void freeComplex2D(Complex **c)
{
  if(c!=NULL){
    free(c);
  }
}

/*------------------------------------------------------------------
  1D,2D FFT
------------------------------------------------------------------*/
int fft_1D(int num, int s, Complex *c)
{
  Complex *c0   = NULL;
  Complex *coef = NULL;
  Complex *src, *dst, *swap;
  int *tbl = NULL;
  int i,j,k;
  int e;
  int n,m;
  
  if(num <= 0 || abs(s) != 1){
    //sȈ
    return 1;
  }

  if(num == 1)
    return 0;

  //
  num = check_2spow(num);

  //ꎞ̈̊m
  c0 = mallocComplex1D(num);
  coef = mallocComplex1D(num);
  tbl = malloc(num * sizeof(int));
  if(c0==NULL || coef==NULL || tbl==NULL){
    //̈mێs
    free0(tbl);
    freeComplex1D(coef);
    freeComplex1D(c0);
    return 2;
  }

  //W̌vZ
  make_coeff_tbl(num, s, coef);
  e = make_bit_reverse_tbl(num, tbl); //2^eĂ

  //Rs[񐔂ɂ
  if(e&1){
    memcpy(c0, c, num * sizeof(Complex));
    src = c0;
    dst = c;
  }else{
    src = c;
    dst = c0;
  }
  
  //i
  for(i=0; i<num; i+=2){
    dst[i].r   = src[tbl[i]].r + src[tbl[i+1]].r;
    dst[i].i   = src[tbl[i]].i + src[tbl[i+1]].i;
    dst[i+1].r = src[tbl[i]].r - src[tbl[i+1]].r;
    dst[i+1].i = src[tbl[i]].i - src[tbl[i+1]].i;
  }

  swap = dst;
  dst = src;
  src = swap;

  //[v
  m = 1;
  n = num>>1;
  for(i=1; i<e; i++){
    m += m;
    n >>= 1;
    for(j=0; j<n; j++){
      for(k=0; k<m; k++){
	int l = k+j*m*2;
	dst[l].r   = src[l].r + coef[n*k].r*src[l+m].r - coef[n*k].i*src[l+m].i;
	dst[l].i   = src[l].i + coef[n*k].r*src[l+m].i + coef[n*k].i*src[l+m].r;
	dst[l+m].r = src[l].r + coef[n*k+m*n].r*src[l+m].r - coef[n*k+m*n].i*src[l+m].i;
	dst[l+m].i = src[l].i + coef[n*k+m*n].r*src[l+m].i + coef[n*k+m*n].i*src[l+m].r;
      }
    }
    swap = dst;
    dst = src;
    src = swap;
  }
  
  free0(tbl);
  freeComplex1D(coef);
  freeComplex1D(c0);

  return 0;
}

// Vector Radix Ƃ͎gĂȂĒP fft_1D() 
// 2珈Ă邾łB
int fft_2D(int col, int row, int s, Complex **c)
{
  Complex *c_row   = NULL;
  int i,j,e;
  
  if(col <= 0 || row <= 0 || abs(s) != 1){
    //sȈ
    return 1;
  }

  if(col == 1 && row == 1)
    return 0;

  //
  row = check_2spow(row);
  col = check_2spow(col);

  //ꎞ̈̊m
  c_row = mallocComplex1D(col);
  if(c_row == NULL){
    //̈mێs
    return 2;
  }

  //sij
  for(i=0; i<col; i++){
    e = fft_1D2(row, s, c[i], i, col-1);
    if(e){
      freeComplex1D(c_row);
      return e;
    }
  }

  //icj
  for(i=0; i<row; i++){
    for(j=0; j<col; j++){
      c_row[j] = c[j][i];
    }
    e = fft_1D2(col, s, c_row, i, row-1);
    if(e){
      freeComplex1D(c_row);
      return e;
    }
    for(j=0; j<col; j++){
      c[j][i] = c_row[j];
    }
  }
  
  freeComplex1D(c_row);
  return 0;
}

//fft_2D()Ŏg߂̊֐
// t==0 ŏAt==end ŃN[Abv
static int fft_1D2(int num, int s, Complex *c, int t, int end)
{
  static Complex *c0   = NULL;
  static Complex *coef = NULL;
  Complex *src, *dst, *swap;
  static int *tbl = NULL;
  static int num_now = 0;
  int i,j,k;
  static int e;
  int n,m;
  
  if(num <= 0 || abs(s) != 1){
    //sȈ
    return 1;
  }

  if(num == 1)
    return 0;

  //
  num = check_2spow(num);

  if(0 == t){
    //
    num_now = num;
    free0(tbl);
    freeComplex1D(coef);
    freeComplex1D(c0);
    
    //ꎞ̈̊m
    c0 = mallocComplex1D(num);
    coef = mallocComplex1D(num);
    tbl = malloc(num * sizeof(int));
    if(c0==NULL || coef==NULL || tbl==NULL){
      //̈mێs
      free0(tbl);
      freeComplex1D(coef);
      freeComplex1D(c0);
      tbl = NULL;
      coef = c0 = NULL;
      num_now = 0;
      return 2;
    }
    
    //W̌vZ
    make_coeff_tbl(num, s, coef);
    e = make_bit_reverse_tbl(num, tbl); //2^eĂ
  }else if(num != num_now){
    //Ăяoꍇ
    return 3;
  }
  
  
  //Rs[񐔂ɂ
  if(e&1){
    memcpy(c0, c, num * sizeof(Complex));
    src = c0;
    dst = c;
  }else{
    src = c;
    dst = c0;
  }
  
  //i
  for(i=0; i<num; i+=2){
    dst[i].r   = src[tbl[i]].r + src[tbl[i+1]].r;
    dst[i].i   = src[tbl[i]].i + src[tbl[i+1]].i;
    dst[i+1].r = src[tbl[i]].r - src[tbl[i+1]].r;
    dst[i+1].i = src[tbl[i]].i - src[tbl[i+1]].i;
  }

  swap = dst;
  dst = src;
  src = swap;

  //[v
  m = 1;
  n = num>>1;
  for(i=1; i<e; i++){
    m += m;
    n >>= 1;
    for(j=0; j<n; j++){
      for(k=0; k<m; k++){
	int l = k+j*m*2;
	dst[l].r   = src[l].r + coef[n*k].r*src[l+m].r - coef[n*k].i*src[l+m].i;
	dst[l].i   = src[l].i + coef[n*k].r*src[l+m].i + coef[n*k].i*src[l+m].r;
	dst[l+m].r = src[l].r + coef[n*k+m*n].r*src[l+m].r - coef[n*k+m*n].i*src[l+m].i;
	dst[l+m].i = src[l].i + coef[n*k+m*n].r*src[l+m].i + coef[n*k+m*n].i*src[l+m].r;
      }
    }
    swap = dst;
    dst = src;
    src = swap;
  }

  if(t == end){
    //N[Abv
    free0(tbl);
    freeComplex1D(coef);
    freeComplex1D(c0);
    tbl = NULL;
    coef = c0 = NULL;
    num_now = 0;
  }
  
  return 0;
}

/*------------------------------------------------------------------
  internal utility functions
------------------------------------------------------------------*/
//f[^̕ѕςɎge[u
static int make_bit_reverse_tbl(int n, int *tbl)
{
  int i;
  int e = 0;
  int s = 1;

  tbl[0] = 0;
  for( ; n>1; n>>=1){
    for(i=1; i<s; i++){
      tbl[i] += tbl[i];
    }
    for(i=0; i<s; i++){
      tbl[i+s] = tbl[i] + 1;
    }
    s += s;
    e++;
  }
  
  return e;
}

//o^tCŽW
static void make_coeff_tbl(int n, int sign, Complex *c)
{
  int i;
  double pn = (2.0 * FFT_PI)/n;

  c[0].r = 1;
  c[0].i = 0;

  for(i=1; i<n; ++i){
    c[i].r = (float)cos(pn * i);
    c[i].i = (float)(sign * sin(pn * i));
  }
}

//2̗ݏ悩`FbNAnƓ傫2̗ݏ̐Ԃ
int check_2spow(int n)
{
  int tmp = n & (n-1);

  //n2̗ݏ
  if(!tmp){
    return n;
  }
  
  while(tmp>0){
    n = tmp; //ޔ
    tmp &= (tmp-1);
  }
  
  return n + n;
}  

static void free0(void *ptr)
{
  if(ptr != NULL){
    free(ptr);
  }
}

